"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: 2D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import numpy
import numpy.ma
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from roguewave.wavetheory.lineardispersion import inverse_intrinsic_dispersion_relation
from typing import List, Union
from roguewave.tools import datetime_to_iso_time_string
import scipy.ndimage
from numpy.ma import MaskedArray
import typing


class WaveSpectrum2DInput(WaveSpectrumInput):
    directions: Union[List[float], numpy.ndarray]

class WaveSpectrum2D(WaveSpectrum):
    def __init__(self,
                 wave_spectrum2D_input:WaveSpectrum2DInput
                 ):


        super().__init__(wave_spectrum2D_input)
        self.direction = numpy.array(wave_spectrum2D_input['directions'],
                                     dtype='float64')
        self._directional_difference = self._delta()

        self._frequency_peak_indices = None
        self._direction_peak_indices = None

    @property
    def e(self) -> numpy.array:
        if self._e is None:
            self._e = self._directional_moment('zero', 0, normalized=False)
        return self._e

    @property
    def a1(self) -> numpy.array:
        if self._a1 is None:
            self._a1 = self._directional_moment('a', 1, normalized=True)
        return self._a1

    @property
    def b1(self) -> numpy.array:
        if self._b1 is None:
            self._b1 = self._directional_moment('b', 1, normalized=True)
        return self._b1

    @property
    def a2(self) -> numpy.array:
        if self._a2 is None:
            self._a2 = self._directional_moment('a', 2, normalized=True)
        return self._a2

    @property
    def b2(self) -> numpy.array:
        if self._b2 is None:
            self._b2 = self._directional_moment('b', 2, normalized=True)
        return self._b2


    @WaveSpectrum.variance_density.setter
    def variance_density(self,val:numpy.ndarray):
        val = val.copy()
        mask = val < 0
        self._variance_density = MaskedArray(val,mask=mask,dtype='float64')
        self._update()

    def _update(self):
        super(WaveSpectrum2D, self)._update()
        self._a1 = None
        self._b1 = None
        self._a2 = None
        self._b2 = None
        self._e = None

    def _delta(self):
        angles = self.direction
        forward_diff = (numpy.diff(angles, append=angles[0]) + 180) % 360 - 180
        backward_diff = (numpy.diff(angles,
                                    prepend=angles[-1]) + 180) % 360 - 180
        return (forward_diff + backward_diff) / 2

    def _directional_moment(self, kind='zero', order=0,
                            normalized=True) -> numpy.array:
        delta = self._directional_difference
        if kind == 'a':
            harmonic = numpy.cos(self.radian_direction * order) * delta
        elif kind == 'b':
            harmonic = numpy.sin(self.radian_direction * order) * delta
        elif kind == 'zero':
            harmonic = delta
        else:
            raise Exception('Unknown moment')
        density = self.variance_density.filled(0)

        values = numpy.sum(density * harmonic[None, :], axis=-1)

        if normalized:
            scale = numpy.sum(density * delta[None, :], axis=-1)
            scale[scale == 0] = 1
        else:
            scale = 1
        return values / scale

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin, fmax)
        return numpy.trapz(self.e[range] * self.frequency[range] ** power,
                           self.frequency[range])

    def _create_wave_spectrum_input(self)->WaveSpectrum2DInput:
        data = self.variance_density.filled(-1)
        return WaveSpectrum2DInput(
            frequency=list(self.frequency),
            directions=list(self.direction),
            varianceDensity=data.tolist(),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
        )

    def copy(self)->"WaveSpectrum2D":
        input = WaveSpectrum2DInput(
            frequency=self.frequency.copy(),
            directions=self.direction.copy(),
            varianceDensity=self.variance_density.copy(),
            timestamp=self.timestamp,
            latitude=self.latitude,
            longitude=self.longitude
        )
        return WaveSpectrum2D(input)

    def __add__(self, other:"WaveSpectrum2D")->"WaveSpectrum2D":
        spectrum = self.copy()

        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.array(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) +
            other_dens.filled(0), mask)
        return spectrum

    def __sub__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.array(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) -
            other_dens.filled(0), mask)
        return spectrum

    def __neg__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        spectrum.variance_density = -spectrum.variance_density
        return spectrum

    def extract(self,mask:numpy.ndarray)->"WaveSpectrum2D":
        spectrum = self.copy()
        density = spectrum.variance_density
        density[~mask] = 0
        density = numpy.ma.masked_array( density, mask=~mask)
        spectrum.variance_density = density
        return spectrum

    def peak_wavenumber_east(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)

    def peak_wavenumber_north(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)


def empty_spectrum2D_like(spectrum:WaveSpectrum2D)->WaveSpectrum2D:
    input = WaveSpectrum2DInput(
        frequency=spectrum.frequency.copy(),
        varianceDensity=numpy.zeros_like(spectrum.variance_density),
        timestamp=spectrum.timestamp,
        latitude=spectrum.latitude,
        longitude=spectrum.longitude,
        directions=spectrum.direction.copy()
    )
    return WaveSpectrum2D(input)







"""
Contents: IO

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines that can be used to save and load Spectral data.

Functions:

- `load_spectrum`, load spectral data
- `save_spectrum`, save spectral data
- `_deserialize`, deserialize data
- `_serialize`, serialize data


How To Use This Module
======================
(See the individual functions for details.)

1. No need for direct import, use roguewave ``import roguewave`` or ``from roguewave import ...``.
2. Save data: save_spectrum(data, filename)
3. load data: data = load_spectrum(filename)

"""
from .spectrum1D import WaveSpectrum1D
from .spectrum2D import WaveSpectrum2D
from .wavespectrum import WaveSpectrum
from pandas import DataFrame, read_json
from . import spectrum1D, spectrum2D
from typing import Union, Dict, List
import pickle
import gzip
from gzip import BadGzipFile
from pickle import UnpicklingError
import json
import msgpack
import os

_UNION = Union[
    WaveSpectrum1D,
    WaveSpectrum2D,
    List[WaveSpectrum1D],
    List[WaveSpectrum2D],
    Dict[str, List[WaveSpectrum1D]],
    Dict[str, List[WaveSpectrum2D]],
    Dict[str, List[List[WaveSpectrum2D]]],
    List[List[WaveSpectrum2D]],
    List[List[DataFrame]],
    Dict[int,List[DataFrame]]
]


def load_spectrum(filename: str) -> _UNION:
    """
    Load spectral data as saved by "save_spectrum" from the given file and
    return a (nested) object. The precise format of the output depends on what
    was saved.

    :param filename: path to file to load.
    :return:
        - WaveSpectrum1D
          A single 1D spectrum

        - WaveSpectrum2D
          A single 2D spectrum

        - List[WaveSpectrum1D]
          List of 1D spectra

        - List[WaveSpectrum2D],
          List of 2D spectra.

        - Dict[str, List[WaveSpectrum1D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed 1D spectra

        - Dict[str, List[WaveSpectrum2D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed and reconstructed/computed 2D spectra

        - List[List[WaveSpectrum2D]]]
          List of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

        - Dict[str, List[List[WaveSpectrum2D]]]
          Dictionary with Spotter/Observation name as keys, each containing a list
          of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.
    """
    try:
        with gzip.open(filename, 'rb') as file:
            data = file.read().decode('utf-8')
        return _deserialize(json.loads(data))
    except BadGzipFile as error:
        pass


    try:
        with open(filename,'rb') as file:
            data = pickle.load(file)
        return _deserialize(data)
    except UnpicklingError:
        pass

    try:
        with open(filename, 'rb') as file:
            data = msgpack.load(file)
        return _deserialize(data)
    except Exception as e:
        raise e


def _deserialize(data):
    """
    Convert the JSON representation of (nested) wavespectra back into native
    python representation.
    :param data: serialized data
    :return: native python representation.
    """

    if isinstance(data, (dict)):
        # if the dictionary contains "directions" or "frequency" as keys these
        # are serialized spectral objects/
        if 'directions' in data:
            # if directions -> create a wavespectrum2d. The keys directly map
            # to the constructor arguments.
            return spectrum2D(**data)
        elif 'frequency' in data:
            # if not directions but has frequency -> create a wavespectrum1d.
            # The keys directly map to the constructor arguments.
            return spectrum1D(**data)
        elif 'dataframe' in data:
            df = read_json(data['dataframe'])
            df['timestamp'] = df['timestamp'].apply(lambda x: x.tz_localize('utc'))
            return df #DataFrame.from_dict(data['dataframe'])
        else:
            # Otherwise- this is a nested object where each key represents data
            # at a different Spotter/location. Loop over all keys and call the
            # function recursively.
            output = {}
            for key in data:
                output[key] = _deserialize(data[key])
            return output
    elif isinstance(data, list):
        # This is a nested object where each entry represents data
        # at a different Spotter/location. Loop over all entries and call the
        # function recursively.
        output = []
        for item in data:
            output.append(_deserialize(item))
        return output
    elif data is None:
        return None
    else:
        raise Exception('Cannot convert to json compatible type')


def _serialize(_input):
    """
    Convert nested objects into a form that can be serialized as JSON. The lowest
    level data structure is either  WaveSpectrum1D or WaveSpectrum2D.

    :param _input: see load_spectrum
    :return: object that can be json Serialized
    """

    if isinstance(_input, WaveSpectrum):
        # If the input is of type spectrum convert the spectra into a JSON form
        if isinstance(_input, WaveSpectrum1D):
            return _input._create_wave_spectrum_input()
        elif isinstance(_input, WaveSpectrum2D):
            return _input._create_wave_spectrum_input()
    elif isinstance(_input, (dict)):
        # If the input is of type dict, loop over all the keys, and call
        # function recursively on each of the elements. The output is stored
        # in a dictionary with the same keys- but now with content that can
        # be serialized.
        output = {}
        for key in _input:
            output[key] = _serialize(_input[key])
        return output
    elif isinstance(_input, list):
        # If the input is of type list, loop over all the entries, and call
        # function recursively on each of the elements. The output is stored
        # in a list of the same length and ordering- but now with content that can
        # be serialized.
        output = []
        for item in _input:
            output.append(_serialize(item))
        return output
    elif isinstance(_input, DataFrame):
        return {'dataframe':_input.to_json(date_unit='s')}
    elif _input is None:
        # handle the "no data" case.
        return None
    else:
        # Otherwise raise error
        print(type(_input), _input)
        raise Exception('Cannot convert to json compatible type')


def save_spectrum(_input: _UNION, filename: str, format='json', separate_spotters_into_files=False):
    """
    Save spectral data, possible in nested form as returned by the spectral
    partition/reconstruction/etc. functions. Data is saved in JSON form
    compressed with gzip.

    :param _input:
        - WaveSpectrum1D
          A single 1D spectrum

        - WaveSpectrum2D
          A single 2D spectrum

        - List[WaveSpectrum1D]
          List of 1D spectra

        - List[WaveSpectrum2D],
          List of 2D spectra.

        - Dict[str, List[WaveSpectrum1D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed 1D spectra

        - Dict[str, List[WaveSpectrum2D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed and reconstructed/computed 2D spectra

        - List[List[WaveSpectrum2D]]]
          List of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

        - Dict[str, List[List[WaveSpectrum2D]]]
          Dictionary with Spotter/Observation name as keys, each containing a list
          of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

    :param filename: path to save data.
    :return: None
    """

    def write(filename, _input, format):
        output = _serialize(_input)
        if format == 'json':
            data = json.dumps(output)
            with gzip.open(filename, 'wb') as file:
                file.write(data.encode('utf-8'))
        elif format == 'pickle':
            with open(filename,'wb') as file:
                pickle.dump(output,file)
        elif format == 'msgpack':
            with open(filename, 'wb') as file:
                msgpack.pack(output,file)
        else:
            raise Exception('Unknown output format')


    if separate_spotters_into_files:
        os.makedirs(filename, exist_ok=True)
        for key in _input:
            name = os.path.join(filename, key)
            write( name,_input[key],format )

    else:
        write(filename,_input,format)




from .spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput
from .spectrum1D import WaveSpectrum1D, WaveSpectrum1DInput
from .estimators import spec2d_from_spec1d, spec1d_from_spec2d

from roguewave import logger

def spectrum1D( frequency , varianceDensity ,a1=None,b1=None,a2=None,b2=None, latitude=None, longitude=None, timestamp=None)->WaveSpectrum1D:
    input = WaveSpectrum1DInput(
        frequency=frequency,
        varianceDensity=varianceDensity,
        timestamp=timestamp,
        latitude=latitude,
        longitude=longitude,
        a1=a1,
        b1=b1,
        a2=a2,
        b2=b2
    )
    return WaveSpectrum1D(input)


def spectrum2D( frequency , directions, varianceDensity, latitude=None, longitude=None, timestamp=None)->WaveSpectrum2D:
    input = WaveSpectrum2DInput(
        frequency=frequency,
        varianceDensity=varianceDensity,
        timestamp=timestamp,
        latitude=latitude,
        longitude=longitude,
        directions=directions
    )
    return WaveSpectrum2D(input)


from roguewave.wavespectra.spectrum1D import WaveSpectrum1D, \
    WaveSpectrum1DInput
import typing
from typing import Dict,List,Union, overload
from pysofar.spotter import Spotter, SofarApi
from datetime import datetime
from roguewave.tools import datetime_to_iso_time_string, to_datetime
from datetime import timedelta
from .exceptions import ExceptionNoFrequencyData
from pandas import read_csv, to_numeric
from roguewave import logger
import numpy
import os
from multiprocessing.pool import ThreadPool
from multiprocessing import get_context
from multiprocessing import cpu_count

MAX_LOCAL_LIMIT = 20
MAXIMUM_NUMBER_OF_WORKERS = 40

def _get_spectrum_from_sofar_spotter_api(
        spotter: Spotter,
        start_date: typing.Union[datetime, str] = None,
        end_date: typing.Union[datetime, str] = None,
        limit: int = MAX_LOCAL_LIMIT,
) -> typing.List[WaveSpectrum1D]:
    """
    Grabs the requested spectra for this spotter based on the given keyword arguments

    :param limit: The limit for data to grab. Defaults to 20, For frequency data max of 100 samples at a time.
    :param start_date: ISO 8601 formatted date string. If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string. If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """


    start_date = to_datetime(start_date)
    end_date = to_datetime(end_date)

    json_data = spotter.grab_data(
        limit=limit,
        start_date=datetime_to_iso_time_string(start_date),
        end_date=datetime_to_iso_time_string(end_date),
        include_frequency_data=True,
        include_directional_moments=True)

    if not json_data['frequencyData']:
        raise ExceptionNoFrequencyData(
            f'Spotter {spotter.id} has no spectral data for the requested time range')

    out = []
    for spectrum in json_data['frequencyData']:
        # Load the data into the input object- this is a one-to-one
        # mapping of keys between dictionaries.
        wave_spectrum_input = WaveSpectrum1DInput(**spectrum)
        out.append(WaveSpectrum1D(wave_spectrum_input))

    return out

@overload
def get_spectrum_from_sofar_spotter_api(
        spotter_ids: List[str],
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None
) -> Dict[str, List[WaveSpectrum1D]]: ...

@overload
def get_spectrum_from_sofar_spotter_api(
        spotter_ids: str,
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None
) -> List[WaveSpectrum1D]: ...

def get_spectrum_from_sofar_spotter_api(
        spotter_ids: Union[str,List[str]],
        start_date: Union[datetime, str] = None,
        end_date: Union[datetime, str] = None,
        session: SofarApi=None,
        verbose = False,
        limit=None,
        parallel_download=True
) -> Union[ Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]]:
    """
    Grabs the requested spectra for this spotter based on the given keyword arguments

    :param limit: The limit for data to grab. Defaults to 20, For frequency data max of 100 samples at a time.
    :param start_date: ISO 8601 formatted date string. If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string. If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """


    if not isinstance( spotter_ids, list ):
        spotter_ids = [spotter_ids]
        return_list = False
    else:
        return_list = True

    if session is None:
        session = SofarApi()


    data = {}
    n = 0
    def worker( spotter_id):
        nonlocal n
        logger.info(f'Downloading data for spotter {spotter_id}')
        try:
            data = _download_spectra(spotter_id,session,start_date,end_date,limit,verbose)
        except ExceptionNoFrequencyData as e:
            data = None
        n+=1
        progress = n/len(spotter_ids) * 100
        logger.info( f'spotter: {spotter_id} - progress: {progress:06.2f} %')
        return data

    if parallel_download:
        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
            out = pool.map(worker, spotter_ids)

        for spotter_id, spectra in zip(spotter_ids,out):
            data[spotter_id] = spectra
    else:
        for spotter_id in spotter_ids:
            logger.info( f'Downloading data for spotter {spotter_id}')
            data[spotter_id] = worker(spotter_id)

    if not return_list:
        return data[spotter_ids[0]]
    else:
        return data

def _download_spectra(spotter_id,session,start_date,end_date,limit,verbose):

    spotter = Spotter(spotter_id, spotter_id, session=session)
    _start_date = start_date
    data = []
    while True:
        # We can only download a maximum of 20 spectra at a time; so we need
        # to loop our request. We do not know how many spotters there are
        # in the given timeframe.
        #
        # assumptions:
        #   - spotter api returns a maximum of 20 items per requests
        #   - requests returned start from the requested start data and
        #     with the last entry either being the last entry that fits
        #     in the requested window, or merely the last sample that fits
        #     in the 20 items.
        #   - requests returned are in order

        if limit:
            # If we hit the limit (if given) of spotters requested, break
            if len(data) >= limit:
                break

            # otherwise, update our next request so we don't overrun
            # the limit
            local_limit = min(limit - len(data), MAX_LOCAL_LIMIT)
        else:
            # if no limit is given, just ask for the maximum allowed.
            local_limit = MAX_LOCAL_LIMIT

        try:
            # Try to get the next batch of spectra
            next = _get_spectrum_from_sofar_spotter_api(spotter, _start_date,
                                                        end_date, local_limit)

        except ExceptionNoFrequencyData as e:
            # If no frequency data was returned, we either...
            if not len(data):
                # ...raise an error, if no data was returned previously (no
                # data available at all)...
                raise e
            else:
                # ...or return, assuming that we exhausted the data was
                # available.
                break

        # Add the data to the list
        data += next

        # If we did not receive all data we requested...
        if len(next) < local_limit:
            # , we are done...
            break
        else:
            # ... else we update the startdate to be the timestamp of the last
            # known entry we received plus a second, and use this as the new
            # start.
            _start_date = to_datetime(next[-1].timestamp) + timedelta(
                seconds=1)
    return data


def get_spectrum_from_parser_output(path: str)->typing.List[WaveSpectrum1D]:
    """

    :param path: Path that contains the output from the spotter parser.
    :return: A list of WaveSpectrum1D objects.
    """
    def load_spectral_file(file):
        if os.path.isfile(file):
            data = read_csv(file).apply(to_numeric, errors='coerce')
        else:
            raise FileNotFoundError(file)

        columns = list(data.columns)
        frequencies = [float(x) for x in columns[8:]]
        values = data[columns[8:]].values

        time_tuple = data[columns[0:6]].values
        time = []
        for index in range(time_tuple.shape[0]):
            time.append(
                datetime(year=time_tuple[index, 0], month=time_tuple[index, 1],
                         day=time_tuple[index, 2], hour=time_tuple[index, 3],
                         minute=time_tuple[index, 4],
                         second=time_tuple[index, 5]))

        return {'time': numpy.array(time), 'frequencies': frequencies,
                'values': values}

    files = ['a1', 'b1', 'a2', 'b2', 'Szz']
    data = {}
    for file_type in files:
        file_location = os.path.join(path, file_type+'.csv')
        data[file_type] = load_spectral_file(file_location)

    number_of_spectra = data['Szz']['values'].shape[0]
    spectra = []
    for index in range(0,number_of_spectra):
        input = WaveSpectrum1DInput(frequency=data['Szz']['frequencies'],
                                    varianceDensity=data['Szz']['values'][index,:],
                                    timestamp=data['Szz']['time'][index], latitude=None,
                                    longitude=None, a1=data['a1']['values'][index,:],
                                    b1=data['b1']['values'][index,:],
                                    a2=data['a2']['values'][index,:],
                                    b2=data['b2']['values'][index,:])
        spectra.append(WaveSpectrum1D(input))
    return spectra



"""
Contents: 1D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""

import numpy
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from roguewave.tools import datetime_to_iso_time_string
from typing import List, Union
from datetime import timedelta


class WaveSpectrum1DInput(WaveSpectrumInput):
    a1: Union[List[float], numpy.ndarray]
    b1: Union[List[float], numpy.ndarray]
    a2: Union[List[float], numpy.ndarray]
    b2: Union[List[float], numpy.ndarray]


class WaveSpectrum1D(WaveSpectrum):
    spectral_density_units = 'm**2/Hertz'

    def __init__(self,
                 wave_spectrum1D_input: WaveSpectrum1DInput
                 ):
        super().__init__(wave_spectrum1D_input)
        self._a1 = numpy.array(wave_spectrum1D_input['a1'])
        self._b1 = numpy.array(wave_spectrum1D_input['b1'])
        self._b2 = numpy.array(wave_spectrum1D_input['b2'])
        self._a2 = numpy.array(wave_spectrum1D_input['a2'])
        self._e = numpy.array(wave_spectrum1D_input['varianceDensity'])

    @WaveSpectrum.variance_density.setter
    def variance_density(self,val:numpy.ndarray):
        self._variance_density = val
        self._e = val

    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = (self._range(fmin, fmax)) & numpy.isfinite(self.e)

        return numpy.trapz(
            self.variance_density[range] * self.frequency[range] ** power,
            self.frequency[range])

    def _create_wave_spectrum_input(self) -> WaveSpectrum1DInput:
        return WaveSpectrum1DInput(
            frequency=list(self.frequency),
            varianceDensity=list(self.variance_density),
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
            a1=list(self.a1),
            b1=list(self.b1),
            a2=list(self.a2),
            b2=list(self.b2)
        )

    def copy(self) -> "WaveSpectrum1D":
        input = WaveSpectrum1DInput(
            frequency=self.frequency.copy(),
            varianceDensity=self.variance_density.copy(),
            timestamp=self.timestamp,
            latitude=self.latitude,
            longitude=self.longitude,
            a1=list(self.a1),
            b1=list(self.b1),
            a2=list(self.a2),
            b2=list(self.b2)
        )
        return WaveSpectrum1D(input)

    def __add__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = (spectrum.variance_density +
            other.variance_density)
        return spectrum

    def __sub__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = (spectrum.variance_density -
            other.variance_density)
        return spectrum

    def __neg__(self, other:"WaveSpectrum1D")->"WaveSpectrum1D":
        spectrum = self.copy()
        spectrum.variance_density = -spectrum.variance_density
        return spectrum


def empty_spectrum1D_like(spectrum: WaveSpectrum1D) -> WaveSpectrum1D:
    input = WaveSpectrum1DInput(
        frequency=spectrum.frequency.copy(),
        varianceDensity=numpy.zeros_like(spectrum.variance_density),
        timestamp=spectrum.timestamp,
        latitude=spectrum.latitude,
        longitude=spectrum.longitude,
        a1=list(spectrum.a1),
        b1=list(spectrum.b1),
        a2=list(spectrum.a2),
        b2=list(spectrum.b2)
    )
    return WaveSpectrum1D(input)


